# -*- coding: utf-8 -*-
import argparse
import logging
import logging.config
import sys

from pysubyt.api import Generator, Settings, Sink
from pysubyt.j2.generator import JinjaBasedGenerator
from pysubyt.sinks import SinkFactory
from pysubyt.sources import SourceFactory

log = logging.getLogger(__name__)


def get_arg_parser():
    """
    Defines the arguments to this script by using Python's
    [argparse](https://docs.python.org/3/library/argparse.html)
    """
    parser = argparse.ArgumentParser(
        description="Produces LD triples a Template",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-l",
        "--logconf",
        type=str,
        action="store",
        help="location of the logging config (yml) to use",
    )

    parser.add_argument(
        "-n",
        "--name",
        action="store",
        required=True,
        help="Speficies the name of the template to use",
    )

    parser.add_argument(
        "-s",
        "--set",
        nargs=2,  # each -s should have 2 arguments
        metavar=("KEY", "FILE"),  # meaning/purpose of those arguments
        action="append",  # multiple -s can be combined
        help=(
            "Multiple entries will add different sets "
            'under sets["KEY"] to the templating process'
        ),
    )

    parser.add_argument(
        "-v",
        "--var",
        nargs=2,  # each -v should have 2 arguments
        metavar=("NAME", "VALUE"),  # meaning/purpose of those arguments
        action="append",  # multiple -v can be combined
        help=(
            "Multiple entries will add different named"
            " variables to the templating process"
        ),
    )

    parser.add_argument(
        "-t",
        "--templates",
        metavar="FOLDER",  # meaning of the argument
        action="store",
        default=".",  # local working directory
        help="Passes the context folder holding all the templates",
    )
    parser.add_argument(
        "-i",
        "--input",
        metavar="FILE",  # meaning of the argument
        action="store",
        help=(
            "Specifies the base input set to run over."
            " Shorthand for -s _ FILE"
        ),
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE|PATTERN",  # meaning of the argument
        action="store",
        help="Specifies where to write the output, can use {uritemplate}.",
    )
    parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help=(
            "Force writing output, do not check "
            "if output files already exist."
        ),
    )
    parser.add_argument(
        "-m",
        "--mode",
        metavar=" (no-)it(eration), (no-)ig(norecase), (no-)fl(atten) ",
        action="store",
        help="""Modifies the mode of iteration:
                1. it (default) vs. no-it: apply template for each
                                                iterated row in the input set
                                           vs. apply it only once once for
                                                the complete input set;
                2. ig vs. no-ig: to be implemented;
                3. fl vs. no-fl: to be implemented.""",
    )
    return parser


def make_service(args: argparse.Namespace) -> Generator:
    template_folder = args.templates
    return JinjaBasedGenerator(template_folder)


def make_sources(args: argparse.Namespace) -> dict:
    inputs = dict()
    if args.input is not None:
        inputs["_"] = SourceFactory.make_source(args.input)
    if args.set is not None:
        for [key, file_name] in args.set:
            inputs[key] = SourceFactory.make_source(file_name)
    return inputs


def make_sink(args: argparse.Namespace) -> Sink:
    return SinkFactory.make_sink(args.output, args.force)


def vars_to_dict(vars: list) -> dict:
    if not vars:
        return None
    # else
    return {name: value for [name, value] in vars}


def enable_logging(args: argparse.Namespace):
    if args.logconf is None:
        return
    import yaml  # conditional dependency -- we only need this (for now)

    # when logconf needs to be read

    with open(args.logconf, "r") as yml_logconf:
        logging.config.dictConfig(
            yaml.load(yml_logconf, Loader=yaml.SafeLoader)
        )
    log.info(f"Logging enabled according to config in {args.logconf}")


def main():
    """
    The main entry point to this module.
    """
    args = get_arg_parser().parse_args()

    enable_logging(args)
    service = make_service(args)
    settings = Settings(args.mode)
    vars_dict = vars_to_dict(args.var)
    inputs = make_sources(args)
    sink = make_sink(args)

    try:
        log.debug("service  = %s" % service)
        log.debug("settings = %s" % settings)
        log.debug("variables = %s" % vars_dict)
        log.debug("inputs   = %s" % inputs)
        log.debug("sink     = %s" % sink)

        service.process(args.name, inputs, settings, sink, vars_dict)

        log.debug("processing done")

    except Exception as e:
        errmsg = "pysubyt processing failed due to <%s>" % e
        log.error(errmsg)
        log.exception(e)
        print("*** ERROR *** " + errmsg, file=sys.stderr)

    finally:
        sink.close()


if __name__ == "__main__":
    main()



""" pysubyt

.. module:: pysubyt
    :synopsis: python implementation for Semantic Uplifting by Templates.
        Helps to produces triples out of various datasources
    :noindex:

.. moduleauthor:: Marc Portier <marc.portier@gmail.com>

"""
from pysubyt.api import Generator, Settings, Sink, Source
from pysubyt.j2.generator import JinjaBasedGenerator
from pysubyt.sinks import SinkFactory
from pysubyt.sources import SourceFactory

__all__ = [
    "Sink",
    "Source",
    "Settings",
    "Generator",
    "SourceFactory",
    "SinkFactory",
    "JinjaBasedGenerator",
]



import itertools
import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Callable, Dict

log = logging.getLogger(__name__)


class Sink(ABC):
    """Abstract Interface for sinks"""

    @abstractmethod
    def close(self):
        """Notifies the Sink all writing has been done - allows cleanup"""

    @abstractmethod
    def add(self, part: str, item: dict = None):
        """Writes out the part to the sink

        :param part: the result for a sepcific part that needs to be sinked
        :type part: str
        :param item: the record for which this part was produced
        :type item: dict
        """


class Source(ABC):
    """Abstract Interface for input Sources - mainly context-manager
    for the actual Iterable to be returned"""

    @abstractmethod
    def __enter__(self) -> Iterable:
        """Source context is expected to yield an Iterable"""

    @abstractmethod
    def __exit__(self, *exc):
        """Source context cleanup"""


# TODO make a pandas source in sources.py


class Settings:
    """Embodies all the actual possible modifiers to the process"""

    _scheme: Dict[str, Dict] = {
        "ignorecase": {
            "default": True,
            "description": (
                "Make all keys lowercase so "
                "to ignore case in key references"
            ),
        },
        "flatten": {
            "default": True,
            "description": (
                "Flatten hierarchical strcutures"
                " by making hierarchical key references."
            ),
        },
        "iteration": {
            "default": True,
            "description": (
                "Perform the iteration outside of "
                "the template to avoid looping inside of it"
            ),
        },
    }
    _negation: str = "no-"

    @staticmethod
    def describe() -> str:
        return "\n".join(
            [
                "%30s: %s" % (key, val["description"])
                for (key, val) in Settings._scheme.items()
            ]
        )

    def __init__(self, modifiers: str = None):
        self._values = {
            key: val["default"] for (key, val) in Settings._scheme.items()
        }
        self.load_from_modifiers(modifiers)

    def load_from_modifiers(self, modifiers: str):
        """
        Parses the -m --mode string into actual properties
        (no-)ig(norecase),(no-)fl(atten),(no-)it(eration)
        """
        if modifiers is None:
            return
        # else
        set_parts: Dict[str, bool] = {
            key: False for (key, val) in self._values.items()
        }

        for part in modifiers.split(","):
            val: bool = True
            if part.startswith(Settings._negation):
                val = False
                part = part[len(Settings._negation) :]
            found = [
                key
                for (key, val) in self._values.items()
                if key.startswith(part)
            ]
            assert (
                len(found) == 1
            ), "ambiguous modifier string '%s' matches list: %s" % (
                part,
                found,
            )
            key = found[0]
            assert not set_parts[key], (
                "ambiguous modifier string "
                f"{part} matches key {key} which is already set"
            )
            set_parts[key] = True
            self._values[key] = val

    def as_modifier_str(self) -> str:
        """
        Reproduces the modifier string that declares these settings
        """
        return ",".join(
            [
                "no-" + key if not val else key
                for (key, val) in self._values.items()
            ]
        )

    def __repr__(self) -> str:
        return "Settings('%s')" % self.as_modifier_str()

    def __getattr__(self, key: str) -> bool:
        return self._values[key]

    def __setattr__(self, key: str, val: bool) -> None:
        if key in ["_values"]:  # actual props to be handled through super
            super(Settings, self).__setattr__(key, val)
            return
        # else  --> dynamic props to be stored in self._values
        if key not in Settings._scheme.keys():
            raise KeyError(
                "attribute '%s' not writeable in Settings object" % key
            )
        self._values[key] = val


class ReIterableAccess(dict):
    """Helper class wraps around an existing dict of iterables,
    making it posible to access the members so that they return
    independent object wrappers that insure totally
    independent iterating loops into them
    """

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getitem__(self, key: str) -> Iterable:
        it = dict.__getitem__(self, key)
        # itertools.tee, makes retrieved iterators iterable again
        #   after accessing it from this dict again
        it, cln = itertools.tee(it, 2)
        dict.__setitem__(self, key, it)
        return cln

    def __setitem__(self, key: str, val: Iterable):
        assert isinstance(
            val, Iterable
        ), "This dict only accepts Iterable objects as value"
        dict.__setitem__(self, key, val)

    def __repr__(self):
        dictrepr = dict.__repr__(self)
        return "%s(%s)" % (type(self).__name__, dictrepr)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v


class IteratorsFromSources(dict):
    """
    Helper class managing the context entry of the various sources.
    """

    def __enter__(self):
        iterators = dict()
        for name, source in self.items():
            it = source.__enter__()
            iterators[name] = it
        return ReIterableAccess(iterators)

    def __exit__(self, *exc):
        for name, source in self.items():
            source.__exit__()


class Generator(ABC):
    """Abstract Base Class for the actual generation Service"""

    @abstractmethod
    def make_render_fn(self, template_name: str) -> Callable:
        """Produces the actual render strategy tied to a specific
        templating implementation"""

    def make_processor(
        self,
        template_name: str,
        sets: Dict[str, Iterable],
        settings: Settings,
        sink: Sink,
        vars_dict: dict = None,
    ):
        return Generator.Processor(
            self.make_render_fn(template_name), sets, settings, sink, vars_dict
        )

    class Processor:
        """Rendition proces Manager - controls queue, manages context

        This mainly introduces
          (1) a fifo queue of one item allowing to detect and mark the 'last'
            element of an iteration
          (2) abstracts the actual template rendition to a Callable function
            producing text out of context ``**kvargs``
        """

        def __init__(
            self,
            render_fn: Callable,
            sets: Dict[str, Iterable],
            settings: Settings,
            sink: Sink,
            vars_dict: dict = None,
        ):
            self.render = render_fn
            self.sets = sets
            self.settings = settings
            self.sink = sink
            self.variables = vars_dict if vars_dict is not None else {}
            self.queued_item = None
            self.isFirst = True
            self.isLast = False
            self.index = 0

        def take(self, next_item):
            """Takes next item to be rendered and synced
                (might queue up to next take before actually processing)

            Pushes the previous taken item out,
                and puts the new item in the queue
            """
            assert (
                next_item is not None
            ), "no item to take - use all_taken() for finalization in stead"
            if not self.isFirst or (
                self.isFirst and self.queued_item is not None
            ):  # on first call only queue
                self.push()
            self.queued_item = next_item

        def all_taken(self):
            """Indicates all items have been taken -- finalization

            Will push the last item (already queued)
            """
            self.isLast = True
            self.push()

        def push(self):
            """Actually pushes the item queued"""
            item = self.queued_item
            log.debug("processing item _ = %s" % item)
            part = self.render(
                _=item,
                sets=self.sets,
                ctrl={
                    "isFirst": self.isFirst,
                    "isLast": self.isLast,
                    "index": self.index,
                    "settings": self.settings,
                },
                **self.variables,
            )
            self.sink.add(part, item)
            self.queued_item = None
            self.isFirst = False
            self.index += 1

    def process(
        self,
        template_name: str,
        inputs: Dict[str, Source],
        settings: Settings,
        sink: Sink,
        vars_dict: dict = None,
    ) -> None:
        """Process the records found in the base input and
            write them to the sink.
        Note the base input is expected to be found at inputs['_']

        :param template_name: name of the template to use
        :type template_name: str
        :param input: dict of named Source objects providing content
        :type inputs: Dict[str, Source]
        :param settings: Settings object holding the
        :type settings: Settings
        :param sink: the sink to write result to
        :type sink: Sink
        """
        # convert inputs into sets
        with IteratorsFromSources(inputs) as sets:
            proc = self.make_processor(
                template_name, sets, settings, sink, vars_dict
            )

            if (
                not settings.iteration or "_" not in sets
            ):  # conditions for collection modus
                settings.iteration = False
                proc.all_taken()
            else:  # default modus
                data = sets["_"]
                for item in data:
                    proc.take(item)
                proc.all_taken()



import os
from typing import Callable

from jinja2 import select_autoescape
from pyrdfj2 import Filters, J2RDFSyntaxBuilder

from pysubyt.api import Generator


class JinjaBasedGenerator(Generator):
    """
    Core class for the jinja based LD Templating service.
    """

    def __init__(self, templates_folder: str):
        """
        Builds the generator that produces LD output from datasources

        :param templates_folder: Location of the templates
        """
        if templates_folder is None:
            templates_folder = "."
        self._templates_folder = templates_folder
        ttl_filter = {"ttl": Filters.all()["xsd"]}

        self.syntax_builder = J2RDFSyntaxBuilder(
            templates_folder,
            extra_filters=ttl_filter,
            jinja_env_variables={
                "autoescape": select_autoescape(
                    disabled_extensions=(
                        "ttl",
                        "txt",
                        "ldt",
                        "json",
                        "jsonld",
                    ),
                    default_for_string=True,
                    default=True,
                )
            },
        )

    def __repr__(self):
        abs_folder = os.path.abspath(self._templates_folder)
        return "JinjaBasedGenerator('%s')" % abs_folder

    def make_render_fn(self, template_name: str) -> Callable:
        return self.syntax_builder._get_rdfsyntax_template(
            template_name
        ).render



import glob
import logging
import mimetypes
import os
import re
from typing import Callable

import requests
import validators
from typeguard import check_type

from pysubyt.api import Source

log = logging.getLogger(__name__)


def assert_readable(file_path):
    assert os.path.isfile(file_path), (
        "File to read '%s' does not exists" % file_path
    )
    assert os.access(file_path, os.R_OK), "Can not read '%s'" % file_path


def fname_from_cdisp(cdisp):
    return re.split(r"; ?filename=", cdisp, flags=re.IGNORECASE)


class SourceFactory:
    _instance = None

    def __init__(self):
        self._register = dict()
        self._ext_mime_map = {"csv": "text/csv"}

    @property
    def ext_2_mime(self):
        return self._ext_mime_map

    def _add(self, mime: str, sourceClass: Callable[[str], Source]) -> None:
        assert mime is not False, "mime cannot be empty to register "
        assert sourceClass is not None, "sourceClass must be provided"
        check_type(sourceClass, Callable[[str], Source])

        self._register[mime] = sourceClass

    def _find(self, mime: str):
        assert mime in self._register, (
            "no Source class available for mime '%s'" % mime
        )
        return self._register[mime]

    @staticmethod
    def instance():
        if SourceFactory._instance is None:
            SourceFactory._instance = SourceFactory()
        return SourceFactory._instance

    @staticmethod
    def register(mime: str, sourceClass: Callable[[str], Source]) -> None:
        SourceFactory.instance()._add(mime, sourceClass)

    @staticmethod
    def map(ext: str, mime: str) -> None:
        SourceFactory.instance().ext_2_mime[ext] = mime

    @staticmethod
    def mime_from_url(url: str) -> str:
        # just get the header, no content yet
        response = requests.head(url, allow_redirects=True)
        if response.status_code == 200:
            mime: str = response.info().get_content_type()
            cdhead = response.headers.get("Content-Disposition")
            if mime == "application/octet-stream" and cdhead is not None:
                cd = fname_from_cdisp(cdhead)
                mime = SourceFactory.mime_from_identifier(cd)
        return mime

    @staticmethod
    def mime_from_identifier(identifier: str) -> str:
        ext = identifier.split(".")[-1]
        mime = SourceFactory.instance().ext_2_mime.get(ext)
        log.debug(f"mapping ext '{ext}' to mime '{mime}'")
        if mime is not None:
            return mime
        # else
        return mimetypes.guess_type(identifier)[0]

    @staticmethod
    def make_source(identifier: str) -> Source:
        if validators.url(identifier):
            mime: str = SourceFactory.mime_from_remote(identifier)
            assert False, "TODO remote Source support - see issues #8"

        # else
        if os.path.isdir(identifier):
            source = FolderSource(identifier)
            return source

        # else
        if glob.has_magic(identifier):
            source = GlobSource(identifier)
            return source

        # else
        mime: str = SourceFactory.mime_from_identifier(identifier)
        assert mime is not None, (
            "no valid mime derived from identifier '%s'" % identifier
        )
        sourceClass: Callable[[str], Source] = SourceFactory.instance()._find(
            mime
        )
        source: Source = sourceClass(identifier)

        return source


class FolderSource(Source):
    def __init__(self, folder_path):
        self._folder = os.path.abspath(folder_path)
        self._sourcefiles = sorted(
            list(next(os.walk(self._folder), (None, None, []))[2])
        )
        assert len(self._sourcefiles) > 0, (
            "FolderSource '%s' should have content files." % self._folder
        )
        self._reset()

    def __repr__(self):
        return "%s('%s')" % (type(self), self._folder)

    def _reset(self):
        self._current_source = None
        self._current_iter = None
        self._ix = -1

    def _exitCurrent(self):
        if self._current_source is not None:
            self._current_source.__exit__()

    def _nextSource(self):
        self._exitCurrent()
        self._ix += 1
        if self._ix < len(self._sourcefiles):
            self._current_source = SourceFactory.make_source(
                os.path.join(self._folder, self._sourcefiles[self._ix])
            )
            self._current_iter = self._current_source.__enter__()
        else:
            self._current_source = None
            self._current_iter = None
            raise StopIteration

    def _nextItem(self):
        # proceed to next element, if needed, proceed to next source
        if self._current_source is None:
            self._nextSource()
        try:
            return next(self._current_iter)
        except StopIteration:
            self._nextSource()
            return next(self._current_iter)
        # else
        raise StopIteration

    def __enter__(self):
        class IterProxy:
            def __init__(self, me):
                self._me = me

            def __iter__(self):
                self._me._reset()
                return self

            def __next__(self):
                return self._me._nextItem()

        return IterProxy(self)

    def __exit__(self):
        # exit the current open source
        self._exitCurrent()
        self._reset()


class GlobSource(FolderSource):
    """For now, this class is inheriting from FolderSource. As soon as another
    class appears with similar behavior to FolderSource and GlobSource, we may
    consider to create an abstract class "CollectionSource" and make all of
    theminherit from this abstract class.
    """

    def __init__(self, pattern, pattern_root_dir="."):
        self._folder = pattern_root_dir
        self._sourcefiles = sorted(
            [p for p in glob.glob(pattern) if os.path.isfile(p)]
        )
        assert len(self._sourcefiles) > 0, (
            "GlobSource '%s' should have content files." % self._folder
        )
        self._reset()


try:
    import csv

    class CSVFileSource(Source):
        """
        Source producing iterator over data-set coming from CSV on file
        """

        def __init__(self, csv_file_path):
            assert_readable(csv_file_path)
            self._csv = csv_file_path

        def __enter__(self):
            self._csvfile = open(self._csv, mode="r", encoding="utf-8-sig")
            return csv.DictReader(self._csvfile, delimiter=",")

        def __exit__(self):
            self._csvfile.close()

        def __repr__(self):
            return "CSVFileSource('%s')" % os.path.abspath(self._csv)

    SourceFactory.register("text/csv", CSVFileSource)
    # wrong, yet useful mime for csv:
    SourceFactory.register("application/csv", CSVFileSource)
except ImportError:
    log.warn("Python CSV module not available -- disabling CSV support!")


try:
    import json

    class JsonFileSource(Source):
        """
        Source producing iterator over data-set coming from json on file
        """

        def __init__(self, json_file_path):
            assert_readable(json_file_path)
            self._json = json_file_path

        def __enter__(self):
            # note this is loading everything in memory
            #   -- will not work for large sets
            #   in the future we might need to consider:
            #   (json-stream)[https://pypi.org/project/json-stream/]
            with open(self._json, mode="r", encoding="utf-8-sig") as jsonfile:
                data = json.load(jsonfile)
                # unwrap nested structures
                while isinstance(data, dict) and len(data.keys()) == 1:
                    child = list(data.values())[0]
                    if isinstance(child, dict) and len(child.keys()) == 0:
                        data = [data]
                    else:
                        data = child
                if not isinstance(data, list):
                    data = [data]
            return iter(data)

        def __exit__(self):
            pass

        def __repr__(self):
            return "JsonFileSource('%s')" % os.path.abspath(self._json)

    SourceFactory.register("application/json", JsonFileSource)
except ImportError:
    log.warn("Python JSON module not available -- disabling JSON support!")


try:
    import xmlasdict

    class XMLFileSource(Source):
        """
        Source producing iterator over data-set coming from XML on file
        """

        def __init__(self, xml_file_path):
            assert_readable(xml_file_path)
            self._xml = xml_file_path

        def __enter__(self):
            with open(self._xml, mode="r", encoding="utf-8-sig") as xmlfile:
                xml_str = xmlfile.read()
                xdict = xmlasdict.parse(xml_str)
                # unpack root wrappers
                data = xdict.unpack()

            return iter(data)

        def __exit__(self):
            pass

        def __repr__(self):
            return "XMLFileSource('%s')" % os.path.abspath(self._xml)

    SourceFactory.map("eml", "text/xml")
    SourceFactory.register("text/xml", XMLFileSource)
    # wrong, yet useful mime for xml:
    SourceFactory.register("application/xml", XMLFileSource)
except ImportError:
    log.warn("Python XML module not available -- disabling XML support!")



import os

from uritemplate import URITemplate, variables

from pysubyt.api import Sink


def assert_writable(file_path: str, force_output: bool = False):
    if not force_output:
        assert not os.path.isfile(file_path), (
            "File to write '%s' alread exists" % file_path
        )
    parent_path = os.path.dirname(os.path.abspath(file_path))
    if not os.path.exists(parent_path):
        os.makedirs(parent_path)
    assert os.access(parent_path, os.W_OK), (
        "Can not write to folder '%s' for creating new files" % parent_path
    )


class SinkFactory:
    @staticmethod
    def make_sink(identifier: str, force_output: bool = False) -> Sink:
        if identifier is None:
            return StdOutSink()
        # else:
        if len(variables(identifier)) == 0:  # identifier is not a pattern
            return SingleFileSink(identifier, force_output)
        # else:                                        #identifier is a pattern
        return PatternedFileSink(identifier, force_output)


class StdOutSink(Sink):
    def __init__(self):
        pass

    def __repr__(self):
        return "StdOutSink"

    def close(self):
        pass

    def add(self, part: str, item: dict = None):
        print(part)


class SingleFileSink(Sink):
    def __init__(self, file_path: str, force_output: bool = False):
        assert_writable(file_path, force_output)
        self._file_path = file_path
        self._fopen = open(file_path, "w")
        self._force_output = force_output

    def __repr__(self):
        return "SingleFileSink('%s', %s)" % (
            os.path.abspath(self._file_path),
            self._force_output,
        )

    def close(self):
        self._fopen.close()
        self._fopen = None

    def add(self, part: str, item: dict = None):
        assert self._fopen is not None, "File to Sink to already close()d"
        self._fopen.write(part)


class PatternedFileSink(Sink):
    def __init__(self, name_pattern: str, force_output: bool = False):
        self._name_template = URITemplate(name_pattern)
        self._force_output = force_output

    def __repr__(self):
        return "PatternedFileSink('%s', %s)" % (
            self._name_template.uri,
            self._force_output,
        )

    def close(self):
        pass

    def add(self, part: str, item: dict = None):
        assert item is not None, "No data context available to expand template"
        file_path = self._name_template.expand(item)
        assert_writable(file_path, self._force_output)
        fopen = open(file_path, "w")
        fopen.write(part)
        fopen.close()
